package vn.teko.auth.auth_flutter

import android.app.Activity
import androidx.annotation.NonNull
import io.flutter.embedding.engine.plugins.FlutterPlugin
import io.flutter.embedding.engine.plugins.activity.ActivityAware
import io.flutter.embedding.engine.plugins.activity.ActivityPluginBinding
import vn.teko.android.auth.data.model.IamToken
import vn.teko.android.auth.data.model.IdToken
import vn.teko.android.auth.facebook.loginWithFacebook
import vn.teko.android.auth.google.loginWithGoogle
import vn.teko.android.auth.login.credential.LoginCredential
import vn.teko.android.auth.login.provider.*
import vn.teko.android.core.util.Result
import vn.teko.android.terra.auth.TerraAuth
import vn.teko.android.terra.auth.TerraAuthInterface
import vn.teko.login.core.ui.LoginCallback
import vn.teko.terra.core.android.terra.TerraApp

/** AuthFlutterPlugin */
class AuthFlutterPlugin : FlutterPlugin, AuthFlutter.FlutterAuthApi, ActivityAware {

    private var activity: Activity? = null

    override fun onAttachedToEngine(@NonNull flutterPluginBinding: FlutterPlugin.FlutterPluginBinding) {
        AuthFlutter.FlutterAuthApi.setup(flutterPluginBinding.binaryMessenger, this)
    }

    override fun onDetachedFromEngine(@NonNull binding: FlutterPlugin.FlutterPluginBinding) {
        AuthFlutter.FlutterAuthApi.setup(binding.binaryMessenger, null)
    }

    override fun getInstance(appName: String): AuthFlutter.CommonAuthResult {
        val resultStatus = AuthFlutter.CommonAuthResult()
        try {
            TerraAuth.getInstance(TerraApp.getInstance(appName))
            resultStatus.isSuccess = true
        } catch (error: Throwable) {
            resultStatus.isSuccess = false
            resultStatus.error = error.message
        }

        return resultStatus
    }

    override fun isAuthorized(
        appName: String,
        result: AuthFlutter.Result<AuthFlutter.IsAuthorizedResult>?
    ) {
        val terraAuth = getTerraAuth(appName)

        try {
            terraAuth.isAuthorized { isAuthorized ->
                result?.success(AuthFlutter.IsAuthorizedResult().also {
                    it.isAuthorized = isAuthorized
                    it.isSuccess = true
                })
            }
        } catch (error: Throwable) {
            result?.success(AuthFlutter.IsAuthorizedResult().also {
                it.isSuccess = false
                it.error = error.message
            })
        }
    }

    override fun getCurrentToken(
        appName: String,
        result: AuthFlutter.Result<AuthFlutter.AccessTokenResult>?
    ) {
        val terraAuth = getTerraAuth(appName)

        try {
            terraAuth.getCurrentToken {
                if (it == null) {
                    result?.success(createErrorAccessTokenResult("Missing IAM Token"))
                } else {
                    result?.success(createSuccessAccessTokenResult(it))
                }
            }
        } catch (error: Throwable) {
            result?.success(createErrorAccessTokenResult(error.message))
        }
    }

    override fun refreshToken(
        appName: String,
        result: AuthFlutter.Result<AuthFlutter.AccessTokenResult>?
    ) {
        val terraAuth = getTerraAuth(appName)

        try {
            terraAuth.refreshToken {
                if (it.isFailure()) {
                    result?.success(createErrorAccessTokenResult(it.exception().message))
                } else {
                    result?.success(createSuccessAccessTokenResult(it.get()))
                }
            }
        } catch (error: Throwable) {
            result?.success(createErrorAccessTokenResult(error.message))
        }
    }

    override fun exchangeToken(
        appName: String,
        iamAudience: String,
        result: AuthFlutter.Result<AuthFlutter.IdTokenResult>?
    ) {
        val terraAuth = getTerraAuth(appName)

        try {
            terraAuth.exchangeIdToken(iamAudience) {
                if (it.isFailure()) {
                    result?.success(createErrorIdTokenResult(it.exception().message))
                } else {
                    result?.success(createSuccessIdTokenResult(it.get()))
                }
            }
        } catch (error: Throwable) {
            result?.success(createErrorIdTokenResult(error.message))
        }
    }

    override fun requestOtp(
        appName: String,
        phoneNumber: String,
        result: AuthFlutter.Result<AuthFlutter.CommonAuthResult>?
    ) {
        val terraAuth = getTerraAuth(appName)

        PhoneProvider.requestPhoneOtp(phoneNumber, terraAuth) {
            if (it.isSuccess()) {
                result?.success(createSuccessCommonResult())
            } else {
                result?.success(createErrorCommonResult(it.exception().message))
            }
        }
    }

    override fun loginWithGoogleCredential(
        appName: String,
        credential: AuthFlutter.GoogleCredential,
        result: AuthFlutter.Result<AuthFlutter.CommonAuthResult>?
    ) {
        val googleCredential = GoogleProvider.getLoginCredential(credential.accessToken.orEmpty())
        loginWithCredential(appName, googleCredential, result)
    }

    override fun loginWithFacebookCredential(
        appName: String,
        credential: AuthFlutter.FacebookCredential,
        result: AuthFlutter.Result<AuthFlutter.CommonAuthResult>?
    ) {
        val facebookCredential =
            FacebookProvider.getLoginCredential(credential.accessToken.orEmpty())
        loginWithCredential(appName, facebookCredential, result)
    }

    override fun loginWithAppleCredential(
        appName: String,
        credential: AuthFlutter.AppleCredential,
        result: AuthFlutter.Result<AuthFlutter.CommonAuthResult>?
    ) {
        result?.success(createErrorCommonResult("Apple Login is not available on Android Platform"))
    }

    override fun loginWithUsernamePasswordCredential(
        appName: String,
        credential: AuthFlutter.UsernamePasswordCredential,
        result: AuthFlutter.Result<AuthFlutter.CommonAuthResult>?
    ) {
        val passwordCredential = UsernamePasswordProvider.getLoginCredential(
            credential.username.orEmpty(),
            credential.password.orEmpty()
        )
        loginWithCredential(appName, passwordCredential, result)
    }

    override fun loginWithPhoneOtpCredential(
        appName: String,
        credential: AuthFlutter.PhoneOtpCredential,
        result: AuthFlutter.Result<AuthFlutter.CommonAuthResult>?
    ) {
        val phoneOtpCredential = PhoneProvider.getLoginCredential(
            credential.phoneNumber.orEmpty(),
            credential.otp.orEmpty()
        )
        loginWithCredential(appName, phoneOtpCredential, result)
    }

    override fun loginWithCustomCredential(
        appName: String,
        credential: AuthFlutter.CustomCredential,
        result: AuthFlutter.Result<AuthFlutter.CommonAuthResult>?
    ) {
        val customCredential = CustomTokenProvider.getLoginCredential(
            credential.idToken.orEmpty()
        )
        loginWithCredential(appName, customCredential, result)
    }

    override fun loginWithGoogle(
        appName: String,
        result: AuthFlutter.Result<AuthFlutter.CommonAuthResult>?
    ) {
        activity?.let {
            val callback = object : LoginCallback {
                override fun onIAMLoginStart() {

                }

                override fun onResult(authResult: Result<Unit, Throwable>) {
                    if (authResult.isSuccess()) {
                        result?.success(createSuccessCommonResult())
                    } else {
                        result?.success(createErrorCommonResult(authResult.exception().message))
                    }
                }
            }

            val terraAuth = getTerraAuth(appName)
            terraAuth.loginWithGoogle(activity = it, callback = callback)
        } ?: result?.success(createErrorCommonResult("Unknown Exception"))
    }

    override fun loginWithFacebook(
        appName: String,
        result: AuthFlutter.Result<AuthFlutter.CommonAuthResult>?
    ) {
        activity?.let {
            val callback = object : LoginCallback {
                override fun onIAMLoginStart() {}

                override fun onResult(authResult: Result<Unit, Throwable>) {
                    if (authResult.isSuccess()) {
                        result?.success(createSuccessCommonResult())
                    } else {
                        result?.success(createErrorCommonResult(authResult.exception().message))
                    }
                }
            }

            val terraAuth = getTerraAuth(appName)
            terraAuth.loginWithFacebook(activity = it, callback = callback)
        } ?: result?.success(createErrorCommonResult("Unknown Exception"))
    }

    override fun loginWithApple(
        appName: String,
        result: AuthFlutter.Result<AuthFlutter.CommonAuthResult>?
    ) {
        result?.success(createErrorCommonResult("Apple Login is not available on Android Platform"))
    }

    private fun loginWithCredential(
        appName: String,
        credential: LoginCredential,
        result: AuthFlutter.Result<AuthFlutter.CommonAuthResult>?
    ) {
        val terraAuth = getTerraAuth(appName)

        try {
            terraAuth.loginWithCredential(credential) {
                if (it.isSuccess()) {
                    result?.success(createSuccessCommonResult())
                } else {
                    result?.success(createErrorCommonResult(it.exception().message))
                }
            }

        } catch (error: Throwable) {
            result?.success(createErrorCommonResult(error.message))
        }
    }

    override fun logout(
        appName: String,
        result: AuthFlutter.Result<AuthFlutter.LogoutResult>?
    ) {
        val terraAuth = getTerraAuth(appName)
        val resultStatus = AuthFlutter.LogoutResult()

        try {
            terraAuth.logOut {
                resultStatus.isSuccess = true
                resultStatus.isLoggedOut = it
                result?.success(resultStatus)
            }
        } catch (error: Throwable) {
            resultStatus.isSuccess = false
            resultStatus.error = error.message
            result?.success(resultStatus)
        }
    }

    override fun onAttachedToActivity(binding: ActivityPluginBinding) {
        setActivityFromActivityPluginBinding(binding)
    }

    override fun onDetachedFromActivityForConfigChanges() {
        detachActivity()
    }

    override fun onReattachedToActivityForConfigChanges(binding: ActivityPluginBinding) {
        setActivityFromActivityPluginBinding(binding)
    }

    override fun onDetachedFromActivity() {
        detachActivity()
    }

    private fun setActivityFromActivityPluginBinding(binding: ActivityPluginBinding) {
        activity = binding.activity
    }

    private fun detachActivity() {
        activity = null
    }

    companion object {
        private fun getTerraAuth(appName: String): TerraAuthInterface {
            return TerraAuth.getInstance(TerraApp.getInstance(appName))
        }

        private fun createSuccessCommonResult(): AuthFlutter.CommonAuthResult {
            return AuthFlutter.CommonAuthResult().apply { isSuccess = true }
        }

        private fun createErrorCommonResult(error: String? = null): AuthFlutter.CommonAuthResult {
            return AuthFlutter.CommonAuthResult().apply {
                isSuccess = false
                this.error = error
            }
        }

        private fun createSuccessAccessTokenResult(token: IamToken): AuthFlutter.AccessTokenResult {
            return AuthFlutter.AccessTokenResult().apply {
                isSuccess = true
                iamToken = AuthFlutter.IamToken().also {
                    it.id = token.id.toString()
                    it.accessToken = token.accessToken
                    it.refreshToken = token.refreshToken
                    it.expiresIn = token.expiresIn.toLong()
                    it.tokenType = token.tokenType
                }
            }
        }

        private fun createSuccessIdTokenResult(token: IdToken): AuthFlutter.IdTokenResult {
            return AuthFlutter.IdTokenResult().apply {
                isSuccess = true
                idToken = AuthFlutter.IdToken().also {
                    it.idToken = token.idToken
                    it.accessToken = token.accessToken
                    it.refreshToken = token.refreshToken
                    it.expiresIn = token.expiresIn?.toLong()
                    it.tokenType = token.tokenType
                }
            }
        }

        private fun createErrorAccessTokenResult(error: String?): AuthFlutter.AccessTokenResult {
            return AuthFlutter.AccessTokenResult().apply {
                isSuccess = false
                this.error = error
            }
        }

        private fun createErrorIdTokenResult(error: String?): AuthFlutter.IdTokenResult {
            return AuthFlutter.IdTokenResult().apply {
                isSuccess = false
                this.error = error
            }
        }
    }
}
