import Flutter
import UIKit
import Janus
import JanusGoogle
import JanusApple
import JanusFacebook
import Terra
import AuthenticationServices

public typealias SwiftGoogleCredential = Janus.GoogleCredential
public typealias PigeonGoogleCredential = auth_flutter.GoogleCredential

public typealias SwiftAppleCredential = Janus.AppleCredential
public typealias PigeonAppleCredential = auth_flutter.AppleCredential

public class SwiftAuthFlutterPlugin: NSObject, FlutterPlugin, FlutterAuthApi {
    
    var completion:( (CommonAuthResult?, FlutterError?) -> Void)? = nil
    
    public func getInstanceAppName(_ appName: String, error: AutoreleasingUnsafeMutablePointer<FlutterError?>) -> CommonAuthResult? {
        do {
            let _ = try getAuthCore(appName: appName)
            return createSuccessCommonResult()
        } catch {
            return createErrorCommonResult(error: error.localizedDescription)
        }
    }
    
    public func isAuthorizedAppName(_ appName: String, completion: @escaping (IsAuthorizedResult?, FlutterError?) -> Void) {
        let result = IsAuthorizedResult()
        
        do {
            let authCore = try getAuthCore(appName: appName)
            result.isSuccess = true
            result.isAuthorized = NSNumber(value: authCore.isAuthorized)
            
        } catch {
            result.isSuccess = false
            result.error = error.localizedDescription
        }
        
        completion(result, nil)
    }
    
    public func getCurrentTokenAppName(_ appName: String, completion: @escaping (AccessTokenResult?, FlutterError?) -> Void) {
        do {
            let authCore = try getAuthCore(appName: appName)
            guard let authInfo = authCore.authInfo else {
                completion(createErrorAccessTokenResult(error: "Missing Iam token"), nil)
                return
            }
            completion(createSuccessAccessTokenResult(authInfo: authInfo), nil)
        } catch {
            completion(createErrorAccessTokenResult(error: error.localizedDescription), nil)
        }
    }
    
    
    public func refreshTokenAppName(_ appName: String, completion: @escaping (AccessTokenResult?, FlutterError?) -> Void) {
        do {
            let authCore = try getAuthCore(appName: appName)
            
            authCore.refreshToken { result, credential, error in
                if (result) {
                    guard let authInfo = credential else {
                        completion(createErrorAccessTokenResult(error: "Missing Iam token"), nil)
                        return
                    }
                    completion(createSuccessAccessTokenResult(authInfo: authInfo), nil)
                } else {
                    completion(createErrorAccessTokenResult(error: error?.localizedDescription), nil)
                }
            }
        } catch {
            completion(createErrorAccessTokenResult(error: error.localizedDescription), nil)
        }
    }
    
    public func exchangeTokenAppName(_ appName: String, iamAudience: String, completion: @escaping (IdTokenResult?, FlutterError?) -> Void) {
        do {
            let authCore = try getAuthCore(appName: appName)
            
            authCore.exchangeToken(audience: iamAudience) { result, credential, error in
                if (result) {
                    guard let authInfo = credential else {
                        completion(createErrorIdTokenResult(error: "Missing IdToken"), nil)
                        return
                    }
                    completion(createSuccessIdTokenResult(authInfo: authInfo), nil)
                } else {
                    completion(createErrorIdTokenResult(error: error?.localizedDescription), nil)
                }
            }
        } catch {
            completion(createErrorIdTokenResult(error: error.localizedDescription), nil)
        }
    }
    
    public func requestOtpAppName(_ appName: String, phoneNumber: String, completion: @escaping (CommonAuthResult?, FlutterError?) -> Void) {
        do {
            let terraApp = try TerraApp.getInstance(appName: appName)
            guard let phoneProvider = PhoneAuthProvider.provider(app: terraApp) else {
                completion(createErrorCommonResult(error: "Missing Phone Provider"), nil)
                return
            }
            
            phoneProvider.requestOTP(phone: phoneNumber) { (isSuccess, error) in
                if (isSuccess) {
                    completion(createSuccessCommonResult(), nil)
                } else {
                    completion(createErrorCommonResult(error: error?.description), nil)
                }
            }
        } catch {
            completion(createErrorCommonResult(error: error.localizedDescription), nil)
        }
    }
    
    private func handleCommonLogin(appName: String, credential: AnyAuthCredential, completion: @escaping (CommonAuthResult?, FlutterError?) -> Void) {
        do {
            let authCore = try getAuthCore(appName: appName)
            self.completion = completion
            try authCore.login(credential: credential, delegate: self)
        } catch {
            let error = error
            completion(createErrorCommonResult(error: error.localizedDescription), nil)
        }
    }
    
    public func login(withGoogleCredentialAppName appName: String, credential: PigeonGoogleCredential, completion: @escaping (CommonAuthResult?, FlutterError?) -> Void) {
        let credential = GoogleAuthProvider.credential(accessToken: credential.accessToken ?? "", userName: nil)
        handleCommonLogin(appName: appName, credential: credential, completion: completion)
    }
    
    public func login(withFacebookCredentialAppName appName: String, credential: FacebookCredential, completion: @escaping (CommonAuthResult?, FlutterError?) -> Void) {
        let credential = FacebookAuthProvider.credential(accessToken: credential.accessToken ?? "", userName: nil)
        handleCommonLogin(appName: appName, credential: credential, completion: completion)
    }
    
    public func login(withAppleCredentialAppName appName: String, credential: PigeonAppleCredential, completion: @escaping (CommonAuthResult?, FlutterError?) -> Void) {
        let credential = AppleAuthProvider.credential(accessToken: credential.accessToken ?? "", userName: nil)
        handleCommonLogin(appName: appName, credential: credential, completion: completion)
    }
    
    public func login(withUsernamePasswordCredentialAppName appName: String, credential: UsernamePasswordCredential, completion: @escaping (CommonAuthResult?, FlutterError?) -> Void) {
        let credential = UserPasswordAuthProvider.credential(username: credential.username ?? "", password: credential.password ?? "")
        handleCommonLogin(appName: appName, credential: credential, completion: completion)
    }
    
    public func login(withPhoneOtpCredentialAppName appName: String, credential: PhoneOtpCredential, completion: @escaping (CommonAuthResult?, FlutterError?) -> Void) {
        do {
            let terraApp = try TerraApp.getInstance(appName: appName)
            guard let phoneProvider = PhoneAuthProvider.provider(app: terraApp) else {
                completion(createErrorCommonResult(error: "Missing Phone Provider"), nil)
                return
            }
            
            let credential = phoneProvider.credential(pin: credential.otp ?? "" )
            handleCommonLogin(appName: appName, credential: credential, completion: completion)
        } catch {
            completion(createErrorCommonResult(error: error.localizedDescription), nil)
        }
    }
    
    public func login(withCustomCredentialAppName appName: String, credential: CustomCredential, completion: @escaping (CommonAuthResult?, FlutterError?) -> Void) {
        let credential = CustomAuthSystemAuthProvider.credential(idToken: credential.idToken ?? "")
        handleCommonLogin(appName: appName, credential: credential, completion: completion)
    }
    
    public func login(withGoogleAppName appName: String, completion: @escaping (CommonAuthResult?, FlutterError?) -> Void) {
        do {
            let authCore = try getAuthCore(appName: appName)
            authCore.registerGoogle()
            guard let vc = UIApplication.topViewController() else {
                completion(createErrorCommonResult(error: "Unknown Exception"), nil)
                return
            }
            self.completion = completion
            authCore.loginWithGoogle(presentViewController: vc, delegate: self)
        } catch {
            completion(createErrorCommonResult(error: error.localizedDescription), nil)
        }
    }
    
    public func login(withFacebookAppName appName: String, completion: @escaping (CommonAuthResult?, FlutterError?) -> Void) {
        do {
            let authCore = try getAuthCore(appName: appName)
            guard let vc = UIApplication.topViewController() else {
                completion(createErrorCommonResult(error: "Unknown Exception"), nil)
                return
            }
            
            self.completion = completion
            
            authCore.loginWithFacebook(presentViewController: vc, delegate: self)
        } catch {
            completion(createErrorCommonResult(error: error.localizedDescription), nil)
        }
    }
    
    public func login(withAppleAppName appName: String, completion: @escaping (CommonAuthResult?, FlutterError?) -> Void) {
        do {
            let authCore = try getAuthCore(appName: appName)
            
            guard let vc = UIApplication.topViewController()  else {
                completion(createErrorCommonResult(error: "Unknown Exception"), nil)
                return
            }
            if #available(iOS 13.0, *) {
                self.completion = completion
                authCore.loginWithApple(presentViewController: vc, delegate: self)
            } else {
                completion(createErrorCommonResult(error: "Apple only supports login for device from iOS 13 only at this time. Sorry for this inconvenience"), nil)
            }
            
        } catch {
            completion(createErrorCommonResult(error: error.localizedDescription), nil)
        }
    }
    
    public func logoutAppName(_ appName: String, completion: @escaping (LogoutResult?, FlutterError?) -> Void) {
        let result = LogoutResult()
        
        do {
            let authCore = try getAuthCore(appName: appName)
            authCore.logout()
            result.isSuccess = true
            result.isLoggedOut = true
        } catch {
            result.isSuccess = false
            result.error = error.localizedDescription
        }
        
        completion(result, nil)
    }
    
    public static func register(with registrar: FlutterPluginRegistrar) {
        let flutterApolloApi = SwiftAuthFlutterPlugin()
        FlutterAuthApiSetup(registrar.messenger(), flutterApolloApi)
    }
    
    private func getAuthCore(appName: String) throws -> AuthCore {
        guard let terraAuth =  TerraAuth.getInstance(by: appName) else {
            throw AuthException.getInstanceError
        }
        
        return terraAuth
    }
}

extension SwiftAuthFlutterPlugin: LoginManagerDelegate {
    public func loginManagerDidLoginSuccess(authInfo: IAuthInfo) {
        completion?(createSuccessCommonResult(), nil)
    }
    
    public func loginManagerDidLoginFail(error: AuthError?) {
        completion?(createErrorCommonResult(error: error?.localizedDescription), nil)
    }
}

// MARK: ASAuthorizationControllerPresentationContextProviding
extension UIViewController: ASAuthorizationControllerPresentationContextProviding {
    
    @available(iOS 13.0, *)
    public func presentationAnchor(for controller: ASAuthorizationController) -> ASPresentationAnchor {
        return self.view.window!
    }
    
}

enum AuthException: Error {
    case getInstanceError
}

extension UIApplication {
    class func topViewController(controller: UIViewController? = UIApplication.shared.keyWindow?.rootViewController) -> UIViewController? {
        if let navigationController = controller as? UINavigationController {
            return topViewController(controller: navigationController.visibleViewController)
        }
        if let tabController = controller as? UITabBarController {
            if let selected = tabController.selectedViewController {
                return topViewController(controller: selected)
            }
        }
        if let presented = controller?.presentedViewController {
            return topViewController(controller: presented)
        }
        
        return controller
    }
}
